@@ -15,15 +15,22 @@ use Carp;
 use Digest::Whirlpool;
 use Digest::MD5;
 use Digest::SHA1;
+use File::Glob;
+use Hash::Util qw( lock_hashref unlock_hashref );
 
-our $VERSION = '1.06';
+our $VERSION = '1.08';
 
 our @ISA    = qw( Exporter );
 our @EXPORT = qw(
 
               file_save
               file_load
-              
+
+              file_mtime
+              file_ctime
+              file_atime
+              file_size
+
               dir_path_make
               dir_path_ensure
 
@@ -41,6 +48,9 @@ our @EXPORT = qw(
               hash_load
               
               hash_validate
+              
+              hash_lock_recursive
+              hash_unlock_recursive
 
               str_url_escape 
               str_url_unescape 
@@ -56,6 +66,8 @@ our @EXPORT = qw(
               wp_hex
               md5_hex
               sha1_hex
+              
+              glob_tree
 
             );
 
@@ -94,6 +106,28 @@ sub file_save
 
 ##############################################################################
 
+sub file_mtime
+{
+  return (stat(shift))[9];
+}
+
+sub file_ctime
+{
+  return (stat(shift))[10];
+}
+
+sub file_atime
+{
+  return (stat(shift))[8];
+}
+
+sub file_size
+{
+  return (stat(shift))[7];
+}
+
+##############################################################################
+
 sub dir_path_make
 {
   my $path = shift;
@@ -249,12 +283,13 @@ sub __hash_ulc
   my $ipl = shift;
   
   my $nr = $ipl ? $hr : {};
-  while( my ( $k, $v ) = each %$hr )
+  for my $k ( keys %$hr )
     {
+    my $v = $hr->{ $k };
     my $old_k = $k;
     $k = $uc ? uc( $k ) : lc( $k );
     $nr->{ $k } = $v;
-    delete $nr->{ $old_k } if $ipl and $k ne $old_k;
+    delete $nr->{ $old_k } if ($ipl and $k ne $old_k);
     }
   return $nr;  
 }
@@ -370,6 +405,33 @@ sub hash_validate
 
 ##############################################################################
 
+# handle recursive hashes untill perl 5.22 etc.
+sub hash_lock_recursive
+{
+  my $hr = shift;
+  
+  lock_hashref( $hr );
+  for my $vr ( values %$hr )
+    {
+    next unless ref( $vr ) eq 'HASH';
+    hash_lock_recursive( $vr );
+    }
+}
+
+sub hash_unlock_recursive
+{
+  my $hr = shift;
+  
+  unlock_hashref( $hr );
+  for my $vr ( values %$hr )
+    {
+    next unless ref( $vr ) eq 'HASH';
+    hash_unlock_recursive( $vr );
+    }
+}
+
+##############################################################################
+
 sub perl_package_to_file
 {
   my $s = shift;
@@ -411,6 +473,42 @@ sub sha1_hex
 
 ##############################################################################
 
+sub __glob_tree_tree_walk
+{
+  my $p = shift; # path
+  my $f = shift; # file mask
+  my $r = shift; # result arr-ref
+
+  #print STDERR "DEBUG: __glob_tree_tree_walk: $p -- $f [$p$f]\n";
+
+  push @$r, grep { -e } sort File::Glob::bsd_glob( "$p$f" );
+
+  opendir( my $dir, "${p}." ) or return undef;
+  my @dirs = sort grep { !/^\./ } grep { -d "$p$_" } readdir $dir;
+  closedir( $dir );
+  
+  #print STDERR "DEBUG: __glob_tree_tree_walk: $p -- $f [$p*] dirs: (@dirs)\n\n";
+
+  __glob_tree_tree_walk( "$p$_/", $f, $r ) for @dirs;
+  
+  return 1;
+}
+
+sub glob_tree
+{
+  my @res;
+  for( @_ )
+    {
+    die "glob_tree: invalid argument" unless /^(.*?\/)([^\/]+)$/;
+    my $p = $1;
+    my $f = $2;
+    __glob_tree_tree_walk( $p, $f, \@res );
+    }
+  return @res;
+}
+
+##############################################################################
+
 BEGIN { __url_escapes_init(); }
 INIT  { __url_escapes_init(); }
 
@@ -434,6 +532,13 @@ INIT  { __url_escapes_init(); }
   my $res     = file_save( $file_name, 'file content here' );
   my $content = file_load( $file_name );
 
+  # --------------------------------------------------------------------------
+
+  my $file_modification_time_in_seconds = file_mtime( $file_name );
+  my $file_change_time_in_seconds       = file_ctime( $file_name );
+  my $file_last_access_time_in_seconds  = file_atime( $file_name );
+  my $file_size                         = file_size(  $file_name );
+
   # --------------------------------------------------------------------------
   
   my $res  = dir_path_make( '/path/to/somewhere' ); # create full path with 0700
@@ -511,6 +616,12 @@ INIT  { __url_escapes_init(); }
   my $sha1_hex      = sha1_hex( $data );
   my $md5_hex       = md5_hex( $data );
 
+  # --------------------------------------------------------------------------
+
+  # find all *.txt files in all subdirectories starting from /usr/local
+  # returned files are with full path names
+  my @files = glob_tree( '/usr/local/*.txt' );
+
 =head1 FUNCTIONS
 
 =head2 hash_validate( $data_hr, $validate_hr );
